/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.Osi7Telegram
 *
 * de.bsvrz.kex.tls.osi7.Osi7Telegram is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.Osi7Telegram is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.Osi7Telegram.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.Osi7Telegram.
 *
 *  de.bsvrz.kex.tls.osi7.Osi7Telegram ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.Osi7Telegram wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7;

import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.hexdump.HexDumper;

/**
 * Stellt eine Stellvertreterobjekt für ein Osi7Telegramm zum einfachen, lesenden Zugriff auf die einzelnen Bestandteile
 * (Einzeltelegramme, De-Blöcke) zur Verfügung.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class Osi7Telegram {
// ------------------------------------------------------------------------
// VARIABLENDEKLARATIONEN
// ------------------------------------------------------------------------

// Sonstige ---------------------------------------------------------------
	/**
	 * DebugLogger für Debug-Ausgaben
	 */
	private static final Debug debug = Debug.getLogger();
// Private ----------------------------------------------------------------
	/**
	 * EinzelTelegramme im Osi7Telegramm.
	 */
	private final SingleTelegram[] _singleTelegram;

	/**
	 * Internes Darstellung eines Osi7Telegramms.
	 */
	private TlsData _osi7TelegramData;

	/**
	 * Anzahl der EinzelTelegramme im Osi7Telegramm. Kann gemäß TLS zwischen 1 und 46 liegen.
	 */
	private final int _countSingleTelegram;

	/**
	 * OSI7 KnotenNummer (die ersten drei Byte eines Osi/Telegramms).
	 */
	private final int _osi7Adress;

	/**
	 * Absender des Telegramms (von OSI-3 aus dem Routing ermittelt)
	 */
	private final int _sender;

// ------------------------------------------------------------------------
// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
// ------------------------------------------------------------------------

	/**
	 * Erzeugt eine Instanz eines Osi7Telegramms aus dem übergebenen Bytearray. Zerlegt das Osi7Telegramm zudem in seine
	 * Einzeltelegramme ({@link SingleTelegram}) und DE-Blöcke {@link DeBlock}
	 *
	 * @param sender       Absender des Telegramms
	 * @param tlsData      Byteblock mit den Daten eines empfangenen Osi7Telegramms gemäß TLS.
	 * @param longtelegram true = Das Telegramm ist ein spezielles Langtelegramm, das größer ist als normale Langtelegramm.
	 *                     Ausserdem ist nur ein Einzeltelegramm vorhanden und das Byte für die "Anzahl Einzeltelegramme"
	 *                     und das Byte "Länge Einzeltelegramme" fehlt; false = Normale Telegramme
	 *
	 * @throws IllegalOsi7TelegramStructureException
	 *          wenn der Telegrammaufbau nicht den formalen Anforderungen der TLS genügt.
	 */
	public Osi7Telegram(int sender, byte[] tlsData, boolean longtelegram) throws IllegalOsi7TelegramStructureException {
		if (longtelegram == false) {
			// Die Daten können normal zerlegt werden
			_osi7TelegramData = new TlsData(tlsData);
			_sender = sender;  // Absender des Telegramms
			if (_osi7TelegramData.length() < 9) {
				throw new IllegalOsi7TelegramStructureException("Osi7Telegramm-Header enthält nur " +
				                                                _osi7TelegramData.length() + " Byte < Minimum von 9 Byte!");
			}
			// Osi7-Adresse steht in den ersten drei Byte (Low-High-ExtraHigh-Byte)
			_osi7Adress = _osi7TelegramData.readUnsignedLowHighExtraHighByteAsInt();

			// Anzahl der folgenden EinzelTelegramme steht im vierten Byte
			_countSingleTelegram = _osi7TelegramData.readUnsignedByteAsInt();

			if (_countSingleTelegram < 1 || _countSingleTelegram > 46) {
				throw new IllegalOsi7TelegramStructureException("Osi7Telegramm enthält " + _countSingleTelegram +
				                                                " Einzeltelegramme -> ausserhalb des gültigen " +
				                                                "Bereichs von 1..46 Einzeltelegrammen!");
			}
			_singleTelegram = new SingleTelegram[_countSingleTelegram];

			//Versuchen, die Einzeltelegramme zu extrahieren...
			try {
				for (int i = 0; i < _countSingleTelegram; ++i) {
					int lengthSingleTelegram = _osi7TelegramData.readUnsignedByteAsInt();

					debug.finer("Erzeuge Einzeltelegramm Nr.: " + (i + 1) + " (von "
					            + _countSingleTelegram + "). Länge Einzeltegramm: " + lengthSingleTelegram);

					if (lengthSingleTelegram < 4 || lengthSingleTelegram > 233) {
						throw new IllegalOsi7TelegramStructureException("Einzeltelegramm enthält " + lengthSingleTelegram +
						                                                " Byte -> ausserhalb des gültigen " +
						                                                "Bereichs von 4..233 Byte!");
					}

					TlsData _singleTelegramData = _osi7TelegramData.readNextTlsDataBlock(lengthSingleTelegram);
					// Einzeltelegramm registrieren....
					// Wenn Knotennummer == 0 (nach TLS eigentlich nicht erlaubt), wird Knotennummer über die Absenderadresse
					// ermittelt.
					// false = Es ist kein besonderes Telegramm und kann normal behandelt werden
					_singleTelegram[i] = new SingleTelegram(_singleTelegramData, (_osi7Adress != 0 ? _osi7Adress : _sender), _sender, false);
				}
			}
			catch (Exception e) {
				String msg;
				msg = "Osi7Telegramms kann nicht in Einzeltelegramme zerlegt werden." + debug.NEWLINE;
				msg += "Länge der Einzeltelegramme größer als Länge des Osi7Telegramms." + debug.NEWLINE;
				msg += HexDumper.toString(tlsData) + debug.NEWLINE;
				msg += "Exception: " + e;
				debug.error(msg, e);
			}
		}
		else {
			// Es ist ein spezielles Langtelegramm
			_osi7TelegramData = new TlsData(tlsData);
			_sender = sender;  // Absender des Telegramms

			if (_osi7TelegramData.length() < 7) {
				throw new IllegalOsi7TelegramStructureException("Osi7Telegramm-Header enthält nur " +
				                                                _osi7TelegramData.length() + " Byte < Minimum von 7 Byte!");
			}
			// Osi7-Adresse steht in den ersten drei Byte (Low-High-ExtraHigh-Byte)
			_osi7Adress = _osi7TelegramData.readUnsignedLowHighExtraHighByteAsInt();

			// In diesem speziellen Langtelegramm steht nur ein Einzeltelegramm
			_countSingleTelegram = 1;

			_singleTelegram = new SingleTelegram[_countSingleTelegram];

			//Versuchen, die Einzeltelegramme zu extrahieren...
			try {
				for (int i = 0; i < _countSingleTelegram; ++i) {
					// Die Länge des Einzeltelegramms kann nicht ausgelesen werden, aber sie kann berechnet werden

					// Gesamtlänge - 3 (Knotennummer)

					int lengthSingleTelegram = tlsData.length - 3;

					TlsData _singleTelegramData = _osi7TelegramData.readNextTlsDataBlock(lengthSingleTelegram);
					// Einzeltelegramm registrieren....
					// Wenn Knotennummer == 0 (nach TLS eigentlich nicht erlaubt), wird Knotennummer über die Absenderadresse
					// ermittelt.
					_singleTelegram[i] = new SingleTelegram(_singleTelegramData, (_osi7Adress != 0 ? _osi7Adress : _sender), _sender, true);
				}
			}
			catch (Exception e) {
				String msg;
				msg = "Osi7Telegramms kann nicht in Einzeltelegramme zerlegt werden." + debug.NEWLINE;
				msg += "Länge der Einzeltelegramme größer als Länge des Osi7Telegramms." + debug.NEWLINE;
				msg += HexDumper.toString(tlsData) + debug.NEWLINE;
				msg += "Exception: " + e;
				debug.error(msg, e);
			}
		}
	}

// ------------------------------------------------------------------------
// GETTER / SETTER METHODEN
// ------------------------------------------------------------------------

	/**
	 * Gibt die Anzahl der Einzeltelgramme im Osi7Telegramm zurück.
	 *
	 * @return Anzahl der Einzeltelegramme (1..46) im Osi7Telegramm.
	 */
	public int getCountSingleTelegram() {
		return _countSingleTelegram;
	}

	/**
	 * Liefert die KnotenNummer des Osi7Telegramms in Dezimaldarstellung (0..16777215).
	 *
	 * @return KnotenNummer des Osi7Telegramms in Dezimaldarstellung (0..16777215)
	 */
	public int getOsi7Adress() {
		return _osi7Adress;
	}

	/**
	 * Liefert die KnotenNummer des Osi7Telegramms als Hex-Darstellung (0..ffffff).
	 *
	 * @return KnotenNummer des Osi7Telegramms als Hex-Darstellung (0..ffffff).
	 */
	public String getOsi7AdressHex() {
		return Integer.toHexString(_osi7Adress);
	}

	/**
	 * Liefert die KnotenNummer des Osi7Telegramms in der Darstellung "LocationCode-Distanz" (0-0 .. 65535-255).
	 *
	 * @return KnotenNummer des Osi7Telegramms in der Darstellung "LocationCode-Distanz" (0-0 .. 65535-255).
	 */
	public String getOsi7AdressLocCode() {
		return ((_osi7Adress >> 8) + "-" + (_osi7Adress & 0xff));
	}

	/**
	 * Gibt den Absender des Telegramms zurück.
	 *
	 * @return Absender des Telegramms.
	 */
	public int getSender() {
		return _sender;
	}

	/**
	 * Gibt die Einzeltelgramme des Osi7Telegramms zurück.
	 *
	 * @return Einzeltelegramme des Osi7Telegramms.
	 */
	public SingleTelegram[] getSingleTelegram() {
		return _singleTelegram;
	}

// ------------------------------------------------------------------------
// KANONISCHE METHODEN
// ------------------------------------------------------------------------

	/**
	 * Liefert die Darstellung eines Osi7Telegramms. Beispielausgabe:
	 *
	 * @return Osi7Telegramm dargestellt mit Basisinformationen und Gliederung in EinzelTelegramme und De-Blöcke.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		final String NEWLINE = System.getProperty("line.separator");
		sb.append("Knotennummer [" + this.getOsi7Adress() +
		          " (x" + this.getOsi7AdressHex() + ") (" + this.getOsi7AdressLocCode() + ")]" +
		          " Anzahl Einzeltelegramme:" + getCountSingleTelegram() + " Länge:" + _osi7TelegramData.length());
		sb.append(NEWLINE);
		for (int i = 0; i < getCountSingleTelegram(); ++i) {
			sb.append(this._singleTelegram[i].toString());
		}
		sb.append(HexDumper.toString(_osi7TelegramData.toBytes()));
		return (sb.toString());
	}
}
